/*
** freebsd.c - Low level kernel access functions for FreeBSD 2.x and 3.x
**
** This program is in the public domain and may be used freely by anyone
** who wants to.
**
** $Id: freebsd.c,v 1.7 2001/07/25 18:12:48 odin Exp $
*/

#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <pwd.h>
#include <errno.h>
#include <string.h>
#include <syslog.h>
#include <nlist.h>
#include <kvm.h>

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/sysctl.h>
#include <sys/queue.h>
#include <sys/uio.h>
#include <sys/socketvar.h>
#define KERNEL
#define _KERNEL
#include <sys/file.h>
#undef KERNEL
#undef _KERNEL
#include <sys/user.h>
#include <sys/filedesc.h>
#include <sys/proc.h>

#include <net/if.h>
#include <net/route.h>

#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <netinet/tcp.h>
#include <netinet/ip_var.h>
#include <netinet/tcp_timer.h>
#include <netinet/tcp_var.h>

#include <arpa/inet.h>

#include <oidentd.h>
#include <oidentd_util.h>
#include <oidentd_inet_util.h>

#ifdef INPLOOKUP_SETLOCAL
#	define _HAVE_OLD_INPCB
#endif

struct nlist nl[] = {
#define N_TCB 0
	{ "_tcb", 0, 0, 0, 0 },
#define N_BTEXT 1
	{ "_btext", 0, 0, 0, 0 },
	{ "", 0, 0, 0, 0 }
};

static int getbuf(u_long addr, void *buf, size_t len);

#ifdef _HAVE_OLD_INPCB

static struct socket *getlist4(	struct inpcb *pcbp,
								in_port_t lport,
								in_port_t fport,
								const struct in_addr *laddr,
								const struct in_addr *faddr);

static struct inpcb tcb;

#else

static struct socket *getlist4(	struct inpcbhead *pcbhead,
								in_port_t lport,
								in_port_t fport,
								const struct in_addr *laddr,
								const struct in_addr *faddr);
static struct inpcbhead tcb;

#endif

static kvm_t *kd;

/*
** Open the kernel memory device
*/

int k_open(void) {
	kd = kvm_openfiles(NULL, "/dev/mem", NULL, O_RDONLY, NULL);
	if (kd == NULL) {
		o_log(DEBUG, "Error: k_open: %s", strerror(errno));
		return (-1);
	}

	/*
	** Extract offsets to the needed variables in the kernel
	*/

	if (kvm_nlist(kd, nl) < 0) {
		o_log(DEBUG, "Error: kvm_nlist: %s", strerror(errno));
		return (-1);
	}

	return (0);
}

/*
** Get a piece of kernel memory with error handling.
** Returns 1 if call succeeded, else 0 (zero).
*/

static int getbuf(u_long addr, void *buf, size_t len) {
	if (addr < nl[N_BTEXT].n_value)
		goto bad_addr;

	if (addr >= (u_long) 0xFFC00000)
		goto bad_addr;

	if ((addr + len) < nl[N_BTEXT].n_value)
		goto bad_addr;

	if ((addr + len) >= (u_long) 0xFFC00000)
		goto bad_addr;

	if (kvm_read(kd, addr, buf, len) < 0) {
		o_log(DEBUG, "getbuf: kvm_read(%08x, %d) : %m", addr, len);
		return (-1);
	}

	return (0);

bad_addr:
	o_log(DEBUG, "getbuf: bad address (%08x not in %08x-0xFFC00000)",
		addr, nl[N_BTEXT].n_value);
	return (-1);
}

/*
** Traverse the inpcb list until a match is found.
** Returns NULL if no match.
*/

#ifdef _HAVE_OLD_INPCB

static struct socket *getlist4(	struct inpcb *pcbp,
								in_port_t lport,
								in_port_t fport,
								const struct in_addr *laddr,
								const struct in_addr *faddr)
{
	struct inpcb *head;

	if (pcbp == NULL)
		return (NULL);

	head = pcbp->inp_prev;

	do {
		if (pcbp->inp_faddr.s_addr == faddr->s_addr &&
			pcbp->inp_laddr.s_addr == laddr->s_addr &&
			pcbp->inp_fport == fport &&
			pcbp->inp_lport == lport)
		{
				return (pcbp->inp_socket);
		}
	} while (pcbp->inp_next != head &&
		getbuf((u_long) pcbp->inp_next, pcbp, sizeof(struct inpcb)) != -1);

	return (NULL);
}

#else

static struct socket *getlist4(	struct inpcbhead *pcbhead,
								in_port_t lport,
								in_port_t fport,
								const struct in_addr *laddr,
								const struct in_addr *faddr)
{
	struct inpcb *head, pcbp;

	(void) laddr;

	head = pcbhead->lh_first;
	if (head == NULL)
		return (NULL);

	do {
		if (getbuf((u_long) head, &pcbp, sizeof(struct inpcb)) == -1)
			break;

		if (pcbp.inp_faddr.s_addr == faddr->s_addr &&
			pcbp.inp_fport == fport && pcbp.inp_lport == lport)

			return (pcbp.inp_socket);

		head = pcbp.inp_list.le_next;
	} while (head != NULL);

	return (NULL);
}

#endif

/*
** Return the user number for the connection owner
*/

int get_user4(	in_port_t lport,
				in_port_t fport,
				struct sockaddr_storage *laddr,
				struct sockaddr_storage *faddr)
{
	struct socket *sockp;
	struct kinfo_proc *kp;
	int i, nentries;

	kp = kvm_getprocs(kd, KERN_PROC_ALL, 0, &nentries);
	if (kp == NULL) {
		o_log(DEBUG, "Error: kvm_getprocs: %s", strerror(errno));
		return (-1);
	}

	if (getbuf(nl[N_TCB].n_value, &tcb, sizeof(tcb)) == -1)
		return (-1);

#ifdef _HAVE_OLD_INPCB
	tcb.inp_prev = (struct inpcb *) nl[N_TCB].n_value;
#endif

	sockp = getlist4(&tcb, lport, fport,
				&SIN4(laddr)->sin_addr, &SIN4(faddr)->sin_addr);

	if (sockp == NULL)
		return (-1);

	/*
	** Locate the file descriptor that has the socket in question
	** open so that we can get the 'ucred' information
	*/

	for (i = 0 ; i < nentries ; i++) {
		if (kp[i].kp_proc.p_fd != NULL) {
			int j;
			int ret;
			struct filedesc pfd;
			struct file **ofiles;

			if (getbuf((u_long) kp[i].kp_proc.p_fd, &pfd, sizeof(pfd)) == -1)
				return (-1);

			ofiles = xmalloc(pfd.fd_nfiles * sizeof(struct file *));

			if (getbuf((u_long) pfd.fd_ofiles, ofiles,
					pfd.fd_nfiles * sizeof(struct file *)) == -1)
			{
				free(ofiles);
				return (-1);
			}

			for (j = 0 ; j < pfd.fd_nfiles ; j++) {
				struct file ofile;

				if (ofiles[j] == NULL)
					continue;

				ret = getbuf((u_long) ofiles[j], &ofile, sizeof(struct file));
				if (ret == -1) {
					free(ofiles);
					return (-1);
				}

				if (ofile.f_count == 0)
					continue;

				if (ofile.f_type == DTYPE_SOCKET &&
					(struct socket *) ofile.f_data == sockp)
				{
					struct pcred pc;

					ret = getbuf((u_long) kp[i].kp_proc.p_cred,
							&pc, sizeof(pc));
					if (ret == -1) {
						free(ofiles);
						return (-1);
					}

					free(ofiles);
					return (pc.p_ruid);
				}
			}

			free(ofiles);
		}
	}

	return (-1);
}

#ifdef WANT_IPV6

int get_user6(	in_port_t lport,
				in_port_t fport,
				struct sockaddr_storage *laddr,
				struct sockaddr_storage *faddr)
{
	struct ucred ucred;
	struct sockaddr_in6 sin6[2];
	int len;
	int ret;

	len = sizeof(struct ucred);

	memset(sin6, 0, sizeof(sin6));

	sin6[0].sin6_len = sizeof(struct sockaddr_in6);
	sin6[0].sin6_family = AF_INET6;
	sin6[0].sin6_port = lport;
	memcpy(&sin6[0].sin6_addr, &SIN6(laddr)->sin6_addr,
		sizeof(sin6[0].sin6_addr));

	sin6[1].sin6_len = sizeof(struct sockaddr_in6);
	sin6[1].sin6_family = AF_INET6;
	sin6[1].sin6_port = fport;
	memcpy(&sin6[1].sin6_addr, &SIN6(faddr)->sin6_addr,
		sizeof(sin6[1].sin6_addr));

	ret = sysctlbyname("net.inet6.tcp6.getcred",
			&ucred, &len, sin6, sizeof(sin6));

	if (ret == -1) {
		o_log(DEBUG, "Error: sysctlbyname: %s", strerror(errno));
		return (-1);
	}

	return (ucred.cr_uid);
}

#endif
