/*
** oidentd_net_util.c - oidentd network utility functions.
** Copyright (C) 2001 Ryan McCabe <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
**
** $Id: oidentd_inet_util.c,v 1.6 2001/07/27 20:34:12 odin Exp $
*/

#define _GNU_SOURCE
#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdarg.h>
#include <ctype.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <syslog.h>
#include <netdb.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <oidentd.h>
#include <oidentd_util.h>
#include <oidentd_inet_util.h>
#include <oidentd_options.h>

/*
** Setup a listening socket.
*/

int setup_listen(in_port_t listen_port, struct sockaddr_storage *listen_addr) {
	const int one = 1;
	int listenfd, ret;
	struct sockaddr_storage sl_sin;

#ifdef WANT_IPV6
	listenfd = socket(PF_INET6, SOCK_STREAM, 0);
	sin_setv6(&SIN6(listen_addr)->sin6_addr, &sl_sin);

	if (listenfd == -1) {
		memset(&sl_sin, 0, sizeof(sl_sin));
#endif
		listenfd = socket(PF_INET, SOCK_STREAM, 0);
		sin_setv4(SIN4(listen_addr)->sin_addr.s_addr, &sl_sin);
#ifdef WANT_IPV6
	}
#endif

	if (listenfd == -1) {
		o_log(DEBUG, "Error: socket: %s", strerror(errno));
		return (-1);
	}

	sin_set_port(htons(listen_port), &sl_sin);

	ret = setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one));

	if (ret != 0) {
		o_log(DEBUG, "Error: setsockopt: %s", strerror(errno));
		return (-1);
	}

	ret = bind(listenfd, (struct sockaddr *) &sl_sin,
			sizeof(struct sockaddr_storage));

	if (ret != 0) {
		o_log(DEBUG, "Error: bind: %s", strerror(errno));
		return (-1);
	}

	if (listen(listenfd, 3) != 0) {
		o_log(DEBUG, "Error: listen: %s", strerror(errno));
		return (-1);
	}

	return (listenfd);
}

/*
** Read at most len bytes from a socket, "sock", store in "buf"
*/

ssize_t sock_read(int sock, u_char *buf, size_t len) {
	u_char c;
	size_t i;
	ssize_t ret;

	if (buf == NULL)
		return (-1);

	for (i = 1 ; i < len ; i++) {
		ojnk:
			if ((ret = read(sock, &c, 1)) == 1) {
				*buf++ = c;
				if (c == '\n')
					break;
			} else if (ret == 0) {
				if (i == 1)
					return (0);
				else
					break;
			} else {
				if (errno == EINTR)
					goto ojnk;

				return (-1);
			}
	}

	*buf = '\0';
	return (i);
}

/*
** Write to a socket, deal with interrupted and incomplete writes.  Returns
** the number of characters written to the socket on success, -1 on failure.
*/

ssize_t sock_write(int sock, void *buf, size_t len) {
	ssize_t n, written = 0;

	while (len > 0) {
		n = write(sock, buf, len);
		if (n == -1) {
			if (errno == EINTR)
				continue;
			return (-1);
		}

		written += n;
		len -= n;
		buf = (u_char *) buf + n;
	}

	return (written);
}

/*
** printf-like function that writes to sockets.
*/

#ifdef HAVE_VASPRINTF

int sockprintf(int fd, char *fmt, ...) {
	va_list ap;
	u_char *buf;
	ssize_t ret;

	va_start(ap, fmt);
	ret = vasprintf((char **) &buf, fmt, ap);
	va_end(ap);

	ret = sock_write(fd, buf, ret);
	free(buf);

	return (ret);
}

#else

int sockprintf(int fd, char *fmt, ...) {
	va_list ap;
	u_char buf[4096];

	va_start(ap, fmt);
	vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);

	return (sock_write(fd, buf, strlen(buf)));
}

#endif

/*
** Return the canonical hostname of the given address.
*/

#ifdef HAVE_GETNAMEINFO

int get_hostname(	struct sockaddr_storage *addr,
					u_char *hostbuf,
					size_t len)
{
	int ret;

	ret = getnameinfo((struct sockaddr *) addr, sizeof(struct sockaddr_storage),
					hostbuf, len, NULL, 0, NI_NAMEREQD);

	return (ret);
}

#else

int get_hostname(	struct sockaddr_storage *addr,
					u_char *hostname,
					size_t len)
{
	struct hostent *host;

	host = gethostbyaddr((char *) sin_addr(addr),
		sizeof(struct sockaddr_storage), addr->ss_family);

	if (host == NULL)
		return (-1);

	if (strlen(host->h_name) >= len)
		return (-1);

	xstrncpy(hostname, host->h_name, len);
	return (0);
}

#endif

/*
** Get the port associated with a tcp service name.
*/

int get_port(const u_char *name, in_port_t *port) {
	struct servent *servent;

	servent = getservbyname(name, "tcp");
	if (servent != NULL)
		*port = ntohs(servent->s_port);
	else {
		char *end;
		long temp_port;

		temp_port = strtol(name, &end, 10);

		if (*end != '\0')
			return (-1);

		if (!VALID_PORT(temp_port))
			return (-1);

		*port = temp_port;
	}

	return (0);
}

/*
** Return a 32-bit, network byte ordered ipv4 or ipv6 address.
*/

#ifdef HAVE_GETADDRINFO

int get_addr(const u_char *hostname, struct sockaddr_storage *addr) {
	struct addrinfo *res;
	size_t len;

	if (getaddrinfo(hostname, NULL, NULL, &res) != 0)
		return (-1);

	switch (res->ai_addr->sa_family) {
		case AF_INET:
			len = sizeof(struct sockaddr_in);
			break;
#ifdef WANT_IPV6
		case AF_INET6:
			len = sizeof(struct sockaddr_in6);
			break;
#endif
		default:
			goto out_fail;
	}

	if (len < res->ai_addrlen)
		goto out_fail;

	memcpy(addr, res->ai_addr, res->ai_addrlen);
	freeaddrinfo(res);

	return (0);

out_fail:
	freeaddrinfo(res);
	return (-1);
}

#else

int get_addr(const u_char *hostname, struct sockaddr_storage *addr) {
	struct in_addr in;

	if (inet_aton(hostname, &in) != 0) {
		sin_setv4(in.s_addr, addr);
		return (0);
	} else {
		struct hostent *host = gethostbyname(hostname);

		if (host != NULL) {
			sin_setv4((in_addr_t) host->h_addr, addr);
			return (0);
		}
	}

	return (-1);
}

#endif

/*
** Return string IPv4 or IPv6 address.
*/

#ifdef HAVE_INET_NTOP

inline void get_ip(	struct sockaddr_storage *ss,
					u_char *buf,
					size_t len)
{
	inet_ntop(ss->ss_family, sin_addr(ss), buf, len);
}

#else

inline void get_ip(	struct sockaddr_storage *ss,
					u_char *buf, 
					size_t len)
{
	in_addr_t addr = ntohl(SIN4(ss)->sin_addr.s_addr);

    snprintf(buf, len, "%u.%u.%u.%u",
            ((addr >> 24) & 0xff), ((addr >> 16) & 0xff),
            ((addr >> 8) & 0xff), (addr & 0xff));
}

#endif

/*
** Setup "ss" as an IPv4 sockaddr struct with the address specified by
** "addr"
*/

void sin_setv4(in_addr_t addr, struct sockaddr_storage *ss) {
	memset(ss, 0, sizeof(struct sockaddr_storage));
	ss->ss_family = AF_INET;
	SIN4(ss)->sin_addr.s_addr = addr;
}

#ifdef WANT_IPV6

/*
** Setup "ss" as an IPv6 sockaddr struct with the address specified by
** "sin6"
*/

void sin_setv6(struct in6_addr *sin6, struct sockaddr_storage *ss) {
	memset(ss, 0, sizeof(struct sockaddr_storage));
	ss->ss_family = AF_INET6;
	memcpy(&SIN6(ss)->sin6_addr, sin6, sizeof(struct sockaddr_in6));
}

#endif

/*
** Copies a sockaddr struct.
*/

void sin_copy(struct sockaddr_storage *ss1, const struct sockaddr_storage *ss2) {
	size_t len;

#ifdef WANT_IPV6
	if (ss2->ss_family == AF_INET6)
		len = sizeof(struct sockaddr_in6);
	else
#endif
		len = sizeof(struct sockaddr_in);

	memset(ss1, 0, sizeof(struct sockaddr_storage));
	memcpy(ss1, ss2, len);
}

/*
** Returns the address set in the appropriate
** sockaddr struct.
*/

void *sin_addr(struct sockaddr_storage *ss) {
#ifdef WANT_IPV6
	if (ss->ss_family == AF_INET6)
		return (&SIN6(ss)->sin6_addr);
#endif

	return (&SIN4(ss)->sin_addr);
}

/*
** Returns the port set in the sockaddr struct.
*/

in_port_t sin_port(const struct sockaddr_storage *ss) {
#ifdef WANT_IPV6
	if (ss->ss_family == AF_INET6)
		return (SIN6(ss)->sin6_port);
#endif

	return (SIN4(ss)->sin_port);
}

/*
** Sets the port for the approprite socket family.
*/

void sin_set_port(in_port_t port, struct sockaddr_storage *ss) {
#ifdef WANT_IPV6
	if (ss->ss_family == AF_INET6)
		SIN6(ss)->sin6_port = port;
#endif

	SIN4(ss)->sin_port = port;
}

/*
** Checks whether two addresses are equal.
*/

bool sin_equal(	struct sockaddr_storage *ss1,
				struct sockaddr_storage *ss2)
{
#ifdef WANT_IPV6
	if (ss1->ss_family == AF_INET6) {
		return (IN6_ARE_ADDR_EQUAL(	&SIN6(ss1)->sin6_addr,
									&SIN6(ss2)->sin6_addr));
	}
#endif

	if (SIN4(ss1)->sin_addr.s_addr == SIN4(ss2)->sin_addr.s_addr)
		return (true);

	return (false);
}

#ifdef WANT_IPV6

/*
** Converts an IPv6-mapped IPv4 address to an IPv4 address.
*/

inline void sin_extractv4(struct in6_addr *in6, struct in_addr *in4) {
	/* XXX - Is there a cleaner portable way to do this? */

	memcpy(in4, ((char *) in6) + 12, sizeof(struct in_addr));
}

#endif
