/*
** oidentd - ojnk ident daemon
** Copyright (C) 1998-2001 Ryan McCabe <odin@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
**
** $Id: oidentd_masq.c,v 1.7 2001/07/27 20:34:12 odin Exp $
*/

#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <setjmp.h>
#include <syslog.h>
#include <string.h>
#include <errno.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <oidentd.h>
#include <oidentd_util.h>
#include <oidentd_inet_util.h>
#include <oidentd_masq.h>
#include <oidentd_options.h>

#ifdef MASQ_SUPPORT

static sigjmp_buf timebuf;
static int fsock;

in_port_t fwdport;
in_addr_t proxy;

extern u_char *ret_os;
extern u_int32_t flags;
extern in_port_t fwdport;

static void fwd_alarm(int sig);

/*
** Parse the masquerading map.
** Returns 0 on success, -1 on failure.
*/

int find_entry(	struct sockaddr_storage *host,
				u_char *user,
				size_t user_size,
				u_char *os,
				size_t os_size)
{
	FILE *fp;
	struct sockaddr_storage addr;
	u_int32_t line_num;
	u_char buf[4096];

#ifdef HAVE_LIBUDB
	if (flags & USEUDB) {
		struct udb_ip_user ibuf;
		struct sockaddr_storage hostaddr;
		u_char ipbuf[MAX_IPLEN];

		memcpy(&hostaddr, host, sizeof(hostaddr));

		get_ip(&hostaddr, ipbuf, sizeof(ipbuf));

		o_log(DEBUG, "[%s] UDB lookup...", ipbuf);

		if (udb_ip_get(SIN4(&hostaddr), &ibuf)) {
			get_ip(&hostaddr, ipbuf, sizeof(ipbuf));
			xstrncpy(user, ibuf.username, user_size);
			xstrncpy(os, ret_os, os_size);

			o_log(NORMAL, "Successful UDB lookup: %s : %s", ipbuf, user);
			return (0);
		}
	}
#endif

	fp = fopen(MASQ_MAP, "r");
	if (fp == NULL) {
		o_log(DEBUG, "Error: fopen: %s: %s", MASQ_MAP, strerror(errno));
		return (-1);
	}

	line_num = 0;

	while (fgets(buf, sizeof(buf), fp) != NULL) {
		struct sockaddr_storage stemp;
		u_char *p, *temp;

		++line_num;
		p = strchr(buf, '\n');
		if (p == NULL) {
			o_log(DEBUG, "[%s:%d] Error: Long line", MASQ_MAP, line_num);
			goto failure;
		}

		p = strtok(buf, " \t");
		if (p == NULL) {
			o_log(DEBUG, "[%s:%d] Error: Missing address parameter",
				MASQ_MAP, line_num);

			goto failure;
		}

		temp = strchr(p, '/');
		if (temp != NULL)
			*temp++ = '\0';

		if (get_addr(p, &stemp) == -1) {
			o_log(DEBUG, "[%s:%d] Error: Invalid address: %s",
				MASQ_MAP, line_num, p);

			goto failure;
		}

		sin_copy(&addr, &stemp);

		if (stemp.ss_family == AF_INET && temp != NULL) {
			in_addr_t mask, mask2;
			char *end;

			mask = strtoul(temp, &end, 10);

			if (*end != '\0') {
				if (get_addr(temp, &stemp) == -1) {
					o_log(DEBUG, "[%s:%d] Error: Invalid address: %s",
						MASQ_MAP, line_num, temp);

					goto failure;
				}

				mask2 = SIN4(&stemp)->sin_addr.s_addr;
			} else {
				if (mask < 1 || mask > 31) {
					o_log(DEBUG, "[%s:%d] Error: Invalid mask: %s",
						MASQ_MAP, line_num, p, temp);

					goto failure;
				}

				mask2 = htonl(~((1 << (32 - mask)) - 1));
			}

			SIN4(&addr)->sin_addr.s_addr &= mask2;
			SIN4(host)->sin_addr.s_addr &= mask2;
		}

		if (sin_equal(&addr, host) == false)
			continue;

		p = strtok(NULL, " \t");
		if (p == NULL) {
			o_log(DEBUG, "[%s:%d] Missing user parameter",
				MASQ_MAP, line_num, p);

			goto failure;
		}

		if (strlen(p) >= user_size) {
			o_log(DEBUG, "[%s:%d] Error: username too long (limit is %d)",
				MASQ_MAP, line_num, user_size);

			goto failure;
		}

		xstrncpy(user, p, user_size);

		p = strtok(NULL, " \t");
		if (p == NULL) {
			o_log(DEBUG, "[%s:%d] Error: Missing OS parameter.",
				MASQ_MAP, line_num, p);

			goto failure;
		}

		if (strlen(p) >= os_size) {
			o_log(DEBUG, "[%s:%d] Error: OS name too long (limit is %d)",
				MASQ_MAP, line_num, os_size);

			goto failure;
		}

		xstrncpy(os, p, os_size);

		fclose(fp);
		return (0);
	}

failure:
	fclose(fp);
	return (-1);
}

/*
** Forward an ident request to another machine, return the response to the
** client that has connected to us and requested it.
*/

int fwd_request(	int sock,
					in_port_t rport,
					in_port_t lport,
					in_port_t mport,
					struct sockaddr_storage *mrelay)
{
	struct sockaddr_storage ss;
	u_char ipbuf[MAX_IPLEN];
	u_char user[512];
	u_char buf[1024];

	fsock = socket(mrelay->ss_family, SOCK_STREAM, 0);
	if (fsock == -1) {
		o_log(DEBUG, "Forward error: socket: %s", strerror(errno));
		return (-1);
	}

	sin_copy(&ss, mrelay);
	sin_set_port(htons(fwdport), &ss);

	if (sigsetjmp(timebuf, 1) != 0)
		return (-1);

	signal(SIGALRM, fwd_alarm);

	/*
	** Five seconds should be plenty, seeing as we're forwarding to a machine
	** on a local network.
	*/

	alarm(5);

	if (connect(fsock, (struct sockaddr *) &ss, sizeof(ss)) != 0) {
		o_log(DEBUG, "Error: Forward: connect: %s", strerror(errno));
		close(fsock);
		return (-1);
	}

	if (sockprintf(fsock, "%d , %d\r\n", lport, mport) < 1) {
		o_log(DEBUG, "Error: Forward: write: %s", strerror(errno));
		close(fsock);
		return (-1);
	}

	if (sock_read(fsock, buf, sizeof(buf)) < 1) {
		close(fsock);
		return (-1);
	}

	/*
	** Don't want timeout when finished processing request
	*/

	alarm(0);
	close(fsock);

	get_ip(mrelay, ipbuf, sizeof(ipbuf));

	if (sscanf(buf, "%*d , %*d : USERID :%*[^:]:%511s", user) != 1) {
		o_log(DEBUG, "[%s] Forwarding response error: \"%s\"", ipbuf, buf);
		return (-1);
	}

	sockprintf(sock, "%d , %d : USERID : %s : %s\r\n",
		rport, mport, ret_os, user);

	o_log(NORMAL,
		"[%s] Successful lookup (by forward): %d , %d : %s",
		ipbuf, rport, mport, user);

	return (0);
}

/*
** Handle the timeout of a forward request.
*/

static void fwd_alarm(int sig) {
	o_log(NORMAL, "Forward timed out.");
	close(fsock);
	siglongjmp(timebuf, sig);
}

#else

/*
** Define a stub masq function.
*/

int masq(	int sock,
			in_port_t lport,
			in_port_t fport,
			struct sockaddr_storage *local,
			struct sockaddr_storage *remote)
{
	(void) sock;
	(void) lport;
	(void) fport;
	(void) local;
	(void) remote;

	return (-1);
}

#endif
