/*
** This program is in the public domain and may be used freely by anyone
** who wants to.
**
** OpenBSD support by Slawomir Piotrowski <slawek@telsatgp.com.pl>
** Now it includes masquerading support.
**
** $Id: openbsd.c,v 1.11 2001/07/25 18:12:48 odin Exp $
*/

#include <config.h>

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <kvm.h>
#include <pwd.h>
#include <nlist.h>
#include <limits.h>
#include <syslog.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <net/if.h>
#include <net/route.h>
#include <netinet/ip.h>
#include <netinet/in_pcb.h>
#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <netinet/ip_fil_compat.h>
#include <netinet/ip_fil.h>
#include <netinet/ip_proxy.h>
#include <netinet/ip_nat.h>

#include <oidentd.h>
#include <oidentd_util.h>
#include <oidentd_inet_util.h>
#include <oidentd_masq.h>
#include <oidentd_options.h>

#ifdef MASQ_SUPPORT
extern in_addr_t proxy;
#endif

extern u_int32_t flags;

static int getbuf(u_long addr, void *buf, size_t len);

static struct socket *getlist(	struct inpcbtable *tcbtablep,
								struct inpcbtable *ktcbtablep,
								in_port_t lport,
								in_port_t fport,
								const struct in_addr *laddr,
								const struct in_addr *faddr);

static kvm_t *kd;

static struct nlist nl[] = {
#define N_TCBTABLE 0
	{"_tcbtable"},
#define N_NATLIST 1
	{"_nat_instances"},
	{""}
};

/* taken from Peter Eriksson <pen@lysator.liu.se> */
int k_open(void) {
	char errbuf[_POSIX2_LINE_MAX];

	/*
	** Open the kernel memory device
	*/

	kd = kvm_openfiles(NULL, "/dev/mem", NULL, O_RDONLY, errbuf);

	if (kd == NULL) {
		o_log(DEBUG, "Error: kvm_openfiles: %s", errbuf);
		return (-1);
	}

	/*
	** Extract offsets to the needed variables in the kernel
	*/

	if (kvm_nlist(kd, nl) < 0) {
		o_log(DEBUG, "Error: kvm_nlist: %s", strerror(errno));
		return (-1);
	}

	return (0);
}

/*
** Get a piece of kernel memory with error handling.
** Returns 1 if call succeeded, else 0 (zero).
** taken from Peter Eriksson <pen@lysator.liu.se>
*/

static int getbuf(u_long addr, void *buf, size_t len) {

	if (kvm_read(kd, addr, buf, len) < 0) {
		o_log(DEBUG, "getbuf: kvm_read(%08lx, %d) : %m", addr, len);
		return (-1);
	}

	return (0);
}

/*
** Traverse the inpcb list until a match is found.
** Returns NULL if no match.
*/

static struct socket *getlist(	struct inpcbtable *tcbtablep,
								struct inpcbtable *ktcbtablep,
								in_port_t lport,
								in_port_t fport,
								const struct in_addr *laddr,
								const struct in_addr *faddr)
{
	struct inpcb *kpcbp, pcb;

	if (tcbtablep == NULL)
		return (NULL);

	kpcbp = tcbtablep->inpt_queue.cqh_first;
	while (kpcbp != (struct inpcb *) ktcbtablep) {
		if (getbuf((u_long) kpcbp, &pcb, sizeof(struct inpcb)) == -1)
			break;

#ifdef MASQ_SUPPORT
		if (flags & PROXY) {
			if (faddr->s_addr == proxy &&
				laddr->s_addr != proxy &&
				pcb.inp_fport == fport &&
				pcb.inp_lport == lport)
			{
				return (pcb.inp_socket);
			}

			if (pcb.inp_faddr.s_addr == faddr->s_addr &&
				pcb.inp_laddr.s_addr == laddr->s_addr &&
		 		pcb.inp_fport == fport &&
				pcb.inp_lport == lport)
			{
				return (pcb.inp_socket);
			}
		}
#endif
		if (pcb.inp_faddr.s_addr == faddr->s_addr &&
			pcb.inp_laddr.s_addr == laddr->s_addr &&
			pcb.inp_fport == fport &&
			pcb.inp_lport == lport)
		{
			return (pcb.inp_socket);
		}

		kpcbp = pcb.inp_queue.cqe_next;
	}

	return (NULL);
}

/*
** Return the user number for the connection owner
** taken from Peter Eriksson <pen@lysator.liu.se>
*/

int get_user4(	in_port_t lport,
				in_port_t fport,
				struct sockaddr_storage *laddr,
				struct sockaddr_storage *faddr)
{
	int ret;
	struct socket *sockp, sock;
	struct inpcbtable tcbtable;

	ret = getbuf(nl[N_TCBTABLE].n_value, &tcbtable, sizeof(tcbtable));
	if (ret == -1)
		return (-1);

	sockp = getlist(&tcbtable, (struct inpcbtable *) nl[N_TCBTABLE].n_value,
					lport, fport,
					&SIN4(laddr)->sin_addr, &SIN4(faddr)->sin_addr);

	if (sockp == NULL)
		return (-1);

	if (getbuf((u_long) sockp, &sock, sizeof(sock)) == -1)
		return (-1);

	if (!(sock.so_state & SS_CONNECTOUT))
		return (-1);

	return (sock.so_ruid);
}

#ifdef MASQ_SUPPORT

/*
** Check ident requests for NAT connection
*/

int masq(	int sock,
			in_port_t lport,
			in_port_t fport,
			struct sockaddr_storage *laddr,
			struct sockaddr_storage *faddr)
{
	nat_t *np;
	nat_t nat;
	u_char os[24];
	u_char user[MAX_ULEN];
	struct sockaddr_storage ss;

	/*
	** Only IPv4 is supported right now..
	*/

	if (faddr->ss_family != AF_INET || laddr->ss_family != AF_INET)
		return (-1);

	if (getbuf(nl[N_NATLIST].n_value, &np, sizeof(np)) == -1)
		return (-1);

	for (; np != NULL ; np = nat.nat_next) {
		int ret;

		if (getbuf((u_long) np, &nat, sizeof(nat)) == -1)
			break;

		if (nat.nat_p != IPPROTO_TCP)
			continue;

		if (lport != nat.nat_outport)
			continue;

		if (fport != nat.nat_oport)
			continue;

		if (SIN4(laddr)->sin_addr.s_addr != nat.nat_outip.s_addr)
			continue;

		if (SIN4(faddr)->sin_addr.s_addr != nat.nat_oip.s_addr) {
			if (!(flags & PROXY))
				continue;

			if (SIN4(faddr)->sin_addr.s_addr != SIN4(&proxy)->sin_addr.s_addr)
				continue;

			if (SIN4(laddr)->sin_addr.s_addr == SIN4(&proxy)->sin_addr.s_addr)
				continue;
		}

		lport = htons(lport);
		fport = htons(fport);

		sin_setv4(nat.nat_inip.s_addr, &ss);

		if (flags & FORWARD) {
			ret = fwd_request(sock, fport, lport, nat.nat_inport, &ss);
			if (ret == 0)
				return (0);

			o_log(DEBUG, "Forward to %s (%d %d) failed.",
				inet_ntoa(nat.nat_inip), lport, nat.nat_inport);
		}

		ret = find_entry(&ss, user, sizeof(user), os, sizeof(os));
		if (ret == 0) {
			u_char ipbuf[MAX_IPLEN];

			sockprintf(sock, "%d , %d : USERID : %s : %s\r\n",
				lport, fport, OS(os), user);

			get_ip(faddr, ipbuf, sizeof(ipbuf));

			o_log(NORMAL,
				"[%s] (NAT) Successful lookup (by IP): %d , %d : %s",
				ipbuf, lport, fport, user);

			return (0);
		}
	}

	return (-1);
}

#endif

#ifdef WANT_IPV6

int get_user6(	in_port_t lport,
				in_port_t fport,
				struct sockaddr_storage *laddr,
				struct sockaddr_storage *faddr)
{
	struct tcp_ident_mapping tir;
	struct sockaddr_in6 *fin;
	struct sockaddr_in6 *lin;
	int mib[] = { CTL_NET, PF_INET, IPPROTO_TCP, TCPCTL_IDENT };
	int error = 0;
	size_t i;

	memset(&tir, 0, sizeof(tir));

	fin = (struct sockaddr_in6 *) &tir.faddr;
	fin->sin6_family = AF_INET6;
	fin->sin6_len = sizeof(struct sockaddr_in6);

	if (faddr->ss_len > sizeof(tir.faddr))
		return (-1);

	memcpy(&fin->sin6_addr, SIN6(faddr)->sin6_addr, sizeof(tir.faddr));
	fin->sin6_port = fport;

	lin = (struct sockaddr_in6 *) &tir.laddr;
	lin->sin6_family = AF_INET6;
	lin->sin6_len = sizeof(struct sockaddr_in6);

	if (laddr->ss_len > sizeof(tir.laddr))
		return (-1);

	memcpy(&lin->sin6_addr, SIN6(laddr)->sin6_addr, sizeof(tir.laddr));
	lin->sin6_port = lport;

	i = sizeof(tir);
	error = sysctl(mib, sizeof(mib) / sizeof(int), &tir, &i, NULL, 0);

	if (!error && tir.ruid != -1)
		return (tir.ruid);

	if (error == -1)
		o_log(DEBUG, "sysctl failed (%m)");

	return (-1);
}

#endif
